#ifndef _CONVOLUTION_H_
#define _CONVOLUTION_H_




#include "SampleBase.h"
#include "brookgenfiles/convolution.h"

////////////////////////////////////////////////////////////////////////////////
//!
//! \class  convolution
//!
//! \brief  contains application properties for the sample convolution
//! 
////////////////////////////////////////////////////////////////////////////////

template <class T>
class Convolution : public SampleBase
{
    public:

        Convolution(char* name);
        void allocateBuffer();
        void fillInputBuffer();
        int run();
        void printTimeInfo();
        void verifyResults();
        void comparePerformance();
        ~Convolution();

    private:
        void _convCPU(T* a, T* b, T* c, int m, int k);
        void _callBrookKernel(int width, int height, ::brook::Stream<T> A,
                              ::brook::Stream<T> B, ::brook::Stream<T> C);

        T* _inputA;
        T* _inputB;
        T* _output;
};

////////////////////////////////////////////////////////////////////////////////
//!
//! \brief  Constructor
//! 
////////////////////////////////////////////////////////////////////////////////

template <class T>
Convolution<T>::Convolution(char* name) : SampleBase(name)
{
    _inputA = NULL;
    _inputB = NULL;
    _output = NULL;
}

////////////////////////////////////////////////////////////////////////////////
// cpu version of convolution
////////////////////////////////////////////////////////////////////////////////

template <class T>
void 
Convolution<T>::_convCPU(T* a, T* b, T* c, int Height, int Width)
{

	int m, n, i, j;
	for (n=0; n<Height; n++) {
		for (m=0; m<Width; m++) {
			c[n*Width+m] = 0;
			for(i=0; i<7; i+=1) {
				for(j=0; j<7; j+=1) {			
					if(n-i>=0 && m-j>= 0) {
						c[n*Width+m] += b[i*8+j] * a[(n-i)*(Width)+m-j];			
					}
				}
			}

		}
	}
}

////////////////////////////////////////////////////////////////////////////////
//!
//! \brief  Allocates memory for buffers used by application
//! 
////////////////////////////////////////////////////////////////////////////////

template <class T>
void
Convolution<T>::allocateBuffer()
{
    int Width = info->Width;
    int Height = info->Height;
    
    _inputA = Util::allocateBuffer<T>(Width, Height);
    _inputB = Util::allocateBuffer<T>(8, 7);
    _output = Util::allocateBuffer<T>(Width, Height);
    
    if (!_inputA || !_inputB || !_output)
    {
        fprintf(stderr, "Error: Memory Allocation.\n");
        exit(-1);
    }
}

////////////////////////////////////////////////////////////////////////////////
//!
//! \brief  Fill the input bufers with random values
//! 
////////////////////////////////////////////////////////////////////////////////

template <class T>
void
Convolution<T>::fillInputBuffer()
{
    int Width = info->Width;
    int Height = info->Height;

	Util::fillBuffer<T>(_inputA, Width, Height, 1, 1, Util::RANDOM);
    Util::fillBuffer<T>(_inputB, 8, 7, 1, 1, Util::RANDOM);
}

////////////////////////////////////////////////////////////////////////////////
//!
//! \brief  Print timing information
//! 
////////////////////////////////////////////////////////////////////////////////

template <class T>
void
Convolution<T>::printTimeInfo()
{
    int Width = info->Width;
    int Height = info->Height;

    double gflop = (double)(2.0 * Width * Height * 7 * 7 * info->Iterations)/(double)(1024 * 1024 * 1024);
    double time = timer->GetElapsedTime();

    std::cout << std::setw(8) << "Width" << std::setw(8) << "Height"
                  << std::setw(16) << "Iterations" << std::setw(16) << "GPU Total Time" 
                  << std::setw(16) << "Gflops" << std::endl;

    std::cout << std::setw(8) << info->Width << std::setw(8) << info->Height
                  << std::setw(16) << info->Iterations << std::setw(16) << time
                  << std::setw(16) << gflop / time << std::endl;

    std::cout << std::endl << std::endl;
}

////////////////////////////////////////////////////////////////////////////////
//!
//! \brief  Verify against CPU implementation if requested 
//! 
////////////////////////////////////////////////////////////////////////////////

template <class T>
void
Convolution<T>::verifyResults()
{
    int Width = info->Width;
    int Height = info->Height;
    int res = 0;
	int chkh = 0;
    T *CPU = Util::allocateBuffer<T>(Width, Height);
    
    if (CPU)
    {
        std::cout<<"-e Verify correct output.\n";
        std::cout<<"Performing Matrix Multiplication on CPU ... ";

        _convCPU(_inputA, _inputB, CPU, Height, Width);
        
        std::cout << "Done\n";

		for (chkh=0; chkh<Height; chkh++) {
            res += Util::compareBuffers<T>((float*)CPU+Width*chkh, (float*)_output+(chkh/4+(chkh%4)*(Height/4))*Width, Width);
		}

//		res += Util::compareBuffers<T>(CPU, _output, Width * Height);

		if (0)
		{
			Util::printBuffer(" input, ", _inputA, 0, 0, Width, Height, Width, Height);
			Util::printBuffer(" input, ", _inputB, 0, 0, 8, 7, 8, 7);
			Util::printBuffer(" GPU, ", _output, 0, 0, Width, Height, Width, Height);
			Util::printBuffer(" CPU, ", CPU, 0, 0, Width, Height, Width, Height);
		}
        if (res)
        {
            std::cout << _name << ": Failed!\n\n";
        }
        else
        {
            std::cout << _name << ": Passed!\n\n" ;
        }

        delete [] CPU;
    }
    else
    {
        fprintf(stderr, "Error: Memory Allocation.\n");
        exit(-1);
    }
}

////////////////////////////////////////////////////////////////////////////////
//!
//! \brief  Compare performance with CPU if requested
//! 
////////////////////////////////////////////////////////////////////////////////

template <class T>
void
Convolution<T>::comparePerformance()
{
    int Width = info->Width;
    int Height = info->Height;
    unsigned int i;
    double cpu_time = 0.0;
    double gpu_time = 0.0;
	double k_time = 0.0;
    double gflop = (double)(2.0 * Width * Height * 7 * 7 * info->Iterations)/(double)(1024 * 1024 * 1024);
    T *CPU = Util::allocateBuffer<T>(Width, Height);
    
    gpu_time = timer->GetElapsedTime();
    timer->Reset();
    timer->Start();
    
    if (CPU)
    {
        std::cout << "-p Compare performance with CPU.\n";
        
        // Record CPU Total time
        
        {
            _convCPU(_inputA, _inputB, CPU, Height, Width);
        }
        
        timer->Stop();

        // Print CPU timing and speedup
        cpu_time = timer->GetElapsedTime();
        
        std::cout << std::setw(8) << "Width" << std::setw(8) << "Height"
                  << std::setw(16) << "Iterations" << std::setw(16) << "CPU Total Time"
                  << std::setw(16) << "GPU Total Time" << std::setw(16) << "Gflops"
                  << std::setw(16) << "Speedup" << std::endl;

        std::cout << std::setw(8) << info->Width << std::setw(8) << info->Height
                  << std::setw(16) << info->Iterations << std::setw(16) << cpu_time
                  << std::setw(16) << gpu_time << std::setw(16) << gflop / gpu_time
                  << std::setw(16) << cpu_time*info->Iterations / gpu_time << std::endl;
        
        std::cout << std::endl << std::endl;
        
        delete [] CPU;
    }
    else
    {
        fprintf(stderr, "Error: Memory Allocation.\n");
        exit(-1);
    }
}

////////////////////////////////////////////////////////////////////////////////
//!
//! \brief  invokes the corresponding brook kernel depending on datatype
//! 
////////////////////////////////////////////////////////////////////////////////

template <class T>
void
Convolution<T>::_callBrookKernel(int width, int height, ::brook::Stream<T> A, ::brook::Stream<T> B,
                                   ::brook::Stream<T> C)
{
}

template<>
void
Convolution<float>::_callBrookKernel(int width, int height, ::brook::Stream<float> A, ::brook::Stream<float> B,
                                   ::brook::Stream<float> C)
{
//	do nothing, we will invoke the kernel from run()

}

////////////////////////////////////////////////////////////////////////////////
//!
//! \brief  backend implementation for convolution
//! 
////////////////////////////////////////////////////////////////////////////////

template <class T>
int
Convolution<T>::run()
{
    int Width = info->Width;
    int Height = info->Height;
    int retVal = 0;

	{
		unsigned int inputDim1[] = {Width/4, Height};
		unsigned int inputDim2[] = {2, 7};
        unsigned int outputDim[] = {Width/4, Height/4};

		// we have 4 output, so we will C divide into c0, c1, c2, c3
        ::brook::Stream<float4> A(2, inputDim1);
        ::brook::Stream<float4> B(2, inputDim2);
        ::brook::Stream<float4> C0(2, outputDim);
        ::brook::Stream<float4> C1(2, outputDim);
        ::brook::Stream<float4> C2(2, outputDim);
        ::brook::Stream<float4> C3(2, outputDim);

        // Record GPU Total time
        // warm up
        {
            // Write to stream 
            A.read(_inputA);
            B.read(_inputB);
			if (A.error()||B.error());

            // Run the brook program 
            convolution_gpu_f4_o4((int)Width, (int)Height, A, B, C0, C1, C2, C3);
            
            // Write data back from stream 
			C0.write(_output);
			C1.write(_output+Width*Height/4);
			C2.write(_output+Width*Height*2/4);
			C3.write(_output+Width*Height*3/4);

            // Handle errors if any
			if(C0.error()||C1.error()||C2.error()||C3.error())
            {
                std::cout << "Error occured" << std::endl;
                std::cout << C0.errorLog() << std::endl;
                retVal = -1;
            }
        }


        // Record GPU Total time
        timer->Start();
        
        for (unsigned int i = 0; i < info->Iterations; ++i)
        {
            // Write to stream 
            A.read(_inputA);
            B.read(_inputB);
			if (A.error()||B.error());

            convolution_gpu_f4_o4((int)Width, (int)Height, A, B, C0, C1, C2, C3);
            
            // Write data back from stream 
			C0.write(_output);
			C1.write(_output+Width*Height/4);
			C2.write(_output+Width*Height*2/4);
			C3.write(_output+Width*Height*3/4);

            // Handle errors if any
			if(C0.error()||C1.error()||C2.error()||C3.error())
            {
                std::cout << "Error occured" << std::endl;
                std::cout << C0.errorLog() << std::endl;
                retVal = -1;
            }
        }

        timer->Stop();
	}


    return retVal;
}

////////////////////////////////////////////////////////////////////////////////
//!
//! \brief  Destructor
//! 
////////////////////////////////////////////////////////////////////////////////

template <class T>
Convolution<T>::~Convolution()
{
    if(_inputA != NULL)
    {
        delete [] _inputA;
    }
    
    if(_inputB != NULL)
    {
        delete [] _inputB;
    }

    if(_output != NULL)
    {
        delete [] _output;
    }
}

#endif // _CONVOLUTION_H_
