//###########################################################################//
//             		Convolution for GPU                      				 //
//                                                                           //
//              YI YANG, University of Central Florida                       //
//                            Feb. 2009                                      //
//###########################################################################//
// Disclaimer:                                                               //
// This code is provided on an "AS IS" basis, without warranty. The author   //
// does not have any liability to you or any other person or entity with     //
// respect to any liability, loss, or damage caused or alleged to have been  //
// caused directly or indirectly by this code.                               //
//###########################################################################//
// This project should be put into the "$NVIDIA CUDA SDK\projects"           //
// $NVIDIA CUDA SDK is the directory of the NVIDIA CUDA SDK                  //
// You can run it by follow: convolution.exe -t
//###########################################################################//


// includes, system
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

// includes, project
#include <cutil.h>
#include <cufft.h>
#include <cudaCom.h>
#include <convolution.h>

// includes, kernels
#include <convolution_kernel.cu>


////////////////////////////////////////////////////////////////////////////////
// declaration, forward
void test(int width, int height, int kWidth, int kHeight, CmdInfo* cmdInfo);

// export C interface
extern "C"
void convolutionCPU(float* a, float* b, float* c, int Width, int Height, int k_Width, int k_Height);

extern "C" void runTest( int argc, char** argv);



////////////////////////////////////////////////////////////////////////////////
// Program main
////////////////////////////////////////////////////////////////////////////////
void main( int argc, char** argv)
{
	CmdInfo info;
	parseCommandLine(argc, argv, &info);
    CUT_DEVICE_INIT(1, "");
    if (info.Test) {
		int start = 256;
		for (int i = start; i<4096*2; i*=2)
			test(i, i, 7, 7, &info);
    }
    else {
        test(info.Width, info.Height, info.KWidth, info.KHeight, &info);
    }
    CUT_EXIT(1, "");
}




////////////////////////////////////////////////////////////////////////////////
//! Run a convolution test for CUDA
////////////////////////////////////////////////////////////////////////////////
// width: input width
// height: input height
// kWidth: kernel width
// kHeight: kernel height
void test(int width, int height, int kWidth, int kHeight, CmdInfo* info) {

	unsigned int nsize = width*height;
	unsigned int ksize = kWidth*kHeight;
    unsigned int nMemSize = sizeof( float) * nsize;
    unsigned int kMemSize = sizeof( float) * ksize;

	// allocate host memory
	float* carray = (float*)malloc(nMemSize);
	float* dst = (float*) malloc( nMemSize);
	float* karray = (float*)malloc(kMemSize);
    float* resultArray = (float*) malloc( nMemSize);


	printf( "initalize convolution, width=%d, height=%d, kwidth=%d, kheight=%d\n",
			width, height, kWidth, kHeight);

	// allocate device memory
    float* d_nArray;
    CUDA_SAFE_CALL( cudaMalloc( (void**) &d_nArray, nMemSize));
    float* d_kArray;
    CUDA_SAFE_CALL( cudaMalloc( (void**) &d_kArray, kMemSize));
	// allocate device memory for result
    float* d_resultArray;
    CUDA_SAFE_CALL( cudaMalloc( (void**) &d_resultArray, nMemSize));



    // assign input value
	for(int i = 0; i < width*height; ++i) {
		carray[i] = rand() / (float)RAND_MAX;
	}
	for(int i = 0; i < kWidth*kHeight; ++i) {
		karray[i] = rand() / (float)RAND_MAX;
	}

	// copy host memory to device
	// in gtx8800, there 16 MPs, so we try to allocate 16 thread blocks
	int bwsize = 256;
	int tx = bwsize;
	int ty = 1;
	int bx = (width)/bwsize;
	int bhsize = (height)*bx/16;
	int by = height/bhsize;
	printf("by=%d, bx=%d, ty=%d, tx=%d, bsize=%d\n", by, bx, ty, tx, bx*by);

	float total = 0, kerneltime = 0;
	int loop = 1;


	//warm up
	{
		CUDA_SAFE_CALL( cudaMemcpy( d_nArray, carray, nMemSize,
									cudaMemcpyHostToDevice) );
		CUDA_SAFE_CALL( cudaMemcpy( d_kArray, karray, kMemSize,
									cudaMemcpyHostToDevice) );

		dim3 threads(tx, ty);
		dim3 grid(bx, by);
		convolution_gpu_unroll<<< grid, threads >>>(width, height, kWidth, kHeight, d_nArray, d_kArray, d_resultArray, bwsize, bhsize);

		CUDA_SAFE_CALL( cudaMemcpy( resultArray, d_resultArray, nMemSize,
								cudaMemcpyDeviceToHost) );

	}

	{
		cudaThreadSynchronize();
		timestamp("start transfer");
		CUDA_SAFE_CALL( cudaMemcpy( d_nArray, carray, nMemSize,
									cudaMemcpyHostToDevice) );
		CUDA_SAFE_CALL( cudaMemcpy( d_kArray, karray, kMemSize,
									cudaMemcpyHostToDevice) );
		dim3 threads(tx, ty);
		dim3 grid(bx, by);
		convolution_gpu_unroll<<< grid, threads >>>(width, height, kWidth, kHeight, d_nArray, d_kArray, d_resultArray, bwsize, bhsize);
		CUDA_SAFE_CALL( cudaMemcpy( resultArray, d_resultArray, nMemSize,
								cudaMemcpyDeviceToHost) );
		cudaThreadSynchronize();
		total = timestamp("done transfer");
	}

	{
		CUDA_SAFE_CALL( cudaMemcpy( d_nArray, carray, nMemSize,
									cudaMemcpyHostToDevice) );
		CUDA_SAFE_CALL( cudaMemcpy( d_kArray, karray, kMemSize,
									cudaMemcpyHostToDevice) );
		cudaThreadSynchronize();
		timestamp("start kernel");

		dim3 threads(tx, ty);
		dim3 grid(bx, by);
		convolution_gpu_unroll<<< grid, threads >>>(width, height, kWidth, kHeight, d_nArray, d_kArray, d_resultArray, bwsize, bhsize);
		cudaThreadSynchronize();
		kerneltime = timestamp("done kernel");

		CUDA_SAFE_CALL( cudaMemcpy( resultArray, d_resultArray, nMemSize,
								cudaMemcpyDeviceToHost) );

	}


	printf("width=%d, height=%d, kWidth=%d, kHeight=%d\n", width, height, kWidth, kHeight);
	printf("total performance %f ms, %f GB/s\n", total,  2*width/1000000.0*height*kWidth*kHeight/total);
	printf("total performance %f ms, %f GB/s\n", kerneltime,  2*width/1000000.0*height*kWidth*kHeight/kerneltime);


    // check result
    // custom output handling when no regression test running
    // in this case check if the result is equivalent to the expected soluion


	if (info->Verbose) {
		printf("input\n");
		for (int i=0; i<height; i++) {
			for (int j=0; j<width; j++) {
				printf("%.2f, ", carray[i*width+j]);
			}
			printf("\n");
		}
		printf("kernel\n");
		for (int i=0; i<kHeight; i++) {
			for (int j=0; j<kWidth; j++) {
				printf("%.2f, ", karray[i*kWidth+j]);
			}
			printf("\n");
		}



		printf("gpu output\n");
		for (int i=0; i<height; i++) {
			for (int j=0; j<width; j++) {
				printf("%.2f, ", resultArray[i*width+j]);
			}
			printf("\n");
		}
	}


	convolutionCPU(carray, karray, dst, width, height, kWidth, kHeight);
	CUTBoolean res = cutCompareL2fe(dst, resultArray, nsize, 1e-6f);
	if (info->Verbose) {
		printf("output\n");
		for (int i=0; i<height; i++) {
			for (int j=0; j<width; j++) {
				printf("%.2f, ", dst[i*width+j]);
			}
			printf("\n");
		}
	}
	printf("Test %s \n", (1 == res) ? "PASSED" : "FAILED");
    // cleanup memory

	free(dst);
    free(resultArray);
	free(carray);
	free(karray);
    CUDA_SAFE_CALL(cudaFree(d_nArray));
    CUDA_SAFE_CALL(cudaFree(d_kArray));
    CUDA_SAFE_CALL(cudaFree(d_resultArray));

}
