//###########################################################################//
//             		Convolution for GPU   					                 //
//                                                                           //
//              YI YANG, University of Central Florida                       //
//                            OCT. 2008                                      //
//###########################################################################//
// Disclaimer:                                                               //
// This code is provided on an "AS IS" basis, without warranty. The author   //
// does not have any liability to you or any other person or entity with     //
// respect to any liability, loss, or damage caused or alleged to have been  //
// caused directly or indirectly by this code.                               //
//###########################################################################//
// kernel part for convolution                                               //
//###########################################################################//


#ifndef _CONVOLUTION_KERNEL_H_
#define _CONVOLUTION_KERNEL_H_

#include <stdio.h>
#include <convolution.h>



// load one line kernel and multiply with one line input
__device__ inline float computeline(float* kshared, float* va, float result, int kWidth) {
	result += kshared[0]*va[0];	// broadcast access the kernel data in share memory
	result += kshared[1]*va[1];
	result += kshared[2]*va[2];
	result += kshared[3]*va[3];
	result += kshared[4]*va[4];
	result += kshared[5]*va[5];
	result += kshared[6]*va[6];
	return result;
}

// width: input width
// height: input height
// kWidth: kernel width
// kHeight: kernel height
// indata: input
// kindata: kernel input
// outdata: output
// bwsize:
// bhsize:
// we partition one big input into many small matrix, the size of matrix will be
// (bwsize, bhsize) every block finshi one samll matrix
__global__	void convolution_gpu_unroll(int width, int height, int kWidth, int kHeight,
		float* indata, float* kindata, float* outdata, int bwsize, int bhsize) {
	__shared__ float shared[1024*3+512+2*128];
	__shared__ float kshared[128];

	int bx = blockIdx.x;
	int by = blockIdx.y;

	int tid = threadIdx.x;
	int btid = bx*bwsize+tid;


	float result0 = 0, result1 = 0, result2 = 0, result3 = 0, result4 = 0, result5 = 0, result6 = 0;
	float va[KWIDTH];
	float ov1, ov2;


	#pragma unroll
	for (int i=0; i<KWIDTH; i++) {
		va[i] = 0;
	}

	if (tid<KWIDTH*KHEIGHT) {	// put kernel in share memory
		kshared[tid] = kindata[tid];
	}


	int bstart = bhsize*by;
	int starth = bstart;
	if (by>0) starth = starth-KHEIGHT;			// by>0, we still need load some data for column boundary
	int offset1 = starth*width+btid;			// every thread load one data from input
	int offset2 = starth*width+bx*bwsize-tid-1;	// some thread must load another data from input because of row boundary
	ov1 = indata[offset1];
	if (tid<kWidth) {
		if (bx>0) {								// there are no row boundary when bx=0
			ov2 = indata[offset2];
		}
		else {
			ov2 = 0;
		}
	}



	int maxh = bhsize*(by+1);
	// compute result for output (row tid, by*bhsize to (by+1)*bhsize)
	for (int i=starth; i<maxh; i+=KHEIGHT) {

		#pragma unroll
		for (int tttt=0; tttt<KHEIGHT; tttt++)
		{
			offset1 += width;
			offset2 += width;

			int h = i+tttt;
			if (h==maxh) return;

			// left 16 for the row boundary, we compute bwsize in one block every time, we need put it in share memory
			int addr = tttt*(bwsize+16)+16+tid;
			shared[addr] = ov1;
			va[0] = ov1;
			if (tid<kWidth)
				shared[addr-tid-tid-1] = ov2;
			__syncthreads();

			// load input line from share memory
			va[1] = shared[--addr];
			va[2] = shared[--addr];
			va[3] = shared[--addr];
			va[4] = shared[--addr];
			va[5] = shared[--addr];
			va[6] = shared[--addr];

			// read next row, try to overlap the memory access and computation
			if ((h+1)<height) {	// if we have another row
				ov1 = indata[offset1];
				if (tid<kWidth) {
					if (bx>0) {
						ov2 = indata[offset2];
					}
					else {
						ov2 = 0;
					}
				}
			}

			// we will reuse the input data va to compute with the kernel
			// the result will keep in 7 register

			if (h>=bstart) {
				outdata[h*width+btid] = computeline(kshared, va, result0, kWidth);
			}
			result0 = computeline(kshared+kWidth, va, result1, kWidth);
			result1 = computeline(kshared+kWidth*2, va, result2, kWidth);
			result2 = computeline(kshared+kWidth*3, va, result3, kWidth);
			result3 = computeline(kshared+kWidth*4, va, result4, kWidth);
			result4 = computeline(kshared+kWidth*5, va, result5, kWidth);
			result5 = computeline(kshared+kWidth*6, va, result6, kWidth);
			result6 = 0;
		}
	}

}


#endif // #ifndef _CONVOLUTION_KERNEL_H_
